# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------

# Standard library imports
from subprocess import CalledProcessError

# External imports
from packaging.version import Version as V

# Bokeh imports
from .config import Config
from .system import run
from .ui import failed, passed

__all__ = (
    "check_checkout",
    "check_git",
    "check_issues",
    "check_release_branch",
    "check_repo",
    "check_tags",
    "check_version_order",
)


def check_git(config: Config) -> None:
    try:
        run("which git")
        passed("Command 'git' is available")
    except CalledProcessError:
        failed("Command 'git' is missing")
        config.abort()


def check_repo(config: Config) -> None:
    try:
        run("git status")
    except CalledProcessError:
        failed("Executing outside of a git repository")
        config.abort()

    try:
        remote = run("git config --get remote.origin.url")
        if "bokeh/bokeh" in remote:
            passed("Executing inside the the bokeh/bokeh repository")
        else:
            failed("Executing OUTSIDE the bokeh/bokeh repository")
            config.abort()
    except CalledProcessError:
        failed("Could not determine Git config remote.origin.url")
        config.abort()


def check_checkout(config: Config) -> None:
    try:
        branch = run("git rev-parse --abbrev-ref HEAD").strip()
        if branch == "master":
            passed("Working on master branch")
        else:
            failed(f"NOT working on master branch ({branch!r})")
            config.abort()

        extras = run("git status --porcelain").split("\n")
        extras = [x for x in extras if x != ""]
        if extras:
            failed("Local checkout is NOT clean", extras)
            config.abort()
        else:
            passed("Local checkout is clean")

        try:
            run("git remote update")
            local = run("git rev-parse @")
            remote = run("git rev-parse @{u}")
            base = run("git merge-base @ @{u}")
            if local == remote:
                passed("Checkout is up to date with GitHub")
            else:
                if local == base:
                    status = "NEED TO PULL"
                elif remote == base:
                    status = "NEED TO PUSH"
                else:
                    status = "DIVERGED"
                failed(f"Checkout is NOT up to date with GitHub ({status})")
                config.abort()
        except CalledProcessError:
            failed("Could not check whether local and GitHub are up to date")
            config.abort()

    except CalledProcessError:
        failed("Could not check the checkout state")
        config.abort()


def check_tags(config: Config) -> None:
    try:
        out = run("git for-each-ref --sort=-taggerdate --format '%(tag)' refs/tags")
        tags = [x.strip("'\"") for x in out.split("\n")]

        if config.version in tags:
            failed(f"There is already an existing tag for new version {config.version!r}")
            config.abort()
        else:
            passed(f"New version {config.version!r} does not already have a tag")

        try:
            config.last_any_version = tags[0]
            passed(f"Detected valid last dev/rc/full version {config.last_any_version!r}")
        except ValueError:
            failed(f"Last dev/rc/full version {config.last_any_version!r} is not a valid Bokeh version!")
            config.abort()

        try:
            config.last_full_version = [tag for tag in tags if ("rc" not in tag and "dev" not in tag)][0]
            passed(f"Detected valid last full release version {config.last_full_version!r}")
        except ValueError:
            failed(f"Last full release version {config.last_full_version!r} is not a valid Bokeh version!")
            config.abort()

    except CalledProcessError:
        failed("Could not detect last version tags")
        config.abort()


def check_version_order(config: Config) -> None:
    if V(config.version) > V(config.last_any_version):
        passed(f"New version {config.version!r} is newer than last version {config.last_any_version!r}")
    else:
        failed(f"New version {config.version!r} is NOT newer than last version {config.last_any_version!r}")
        config.abort()


def check_release_branch(config: Config) -> None:
    out = run("git branch --list %s" % config.release_branch)
    if out:
        failed(f"Release branch {config.release_branch!r} ALREADY exists")
        config.abort()
    else:
        passed(f"Release branch {config.release_branch!r} does not already exist")


def check_issues(config: Config) -> None:
    try:
        run(f"python scripts/issues.py -c -p {config.last_full_version}")
        passed("Issue labels are BEP-1 compliant")
    except CalledProcessError as e:
        if "HTTP Error 403: Forbidden" in e.output:
            failed("Issues cannot be checked right now due to GitHub rate limiting")
        else:
            failed("Issue labels are NOT BEP-1 compliant", e.output.split("\n"))
        config.abort()



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------

# Standard library imports
import sys

# Bokeh imports
from .config import Config
from .stages import BUILD_CHECKS, BUILD_STEPS
from .ui import banner, blue

config = Config(sys.argv[1])

banner(blue, "{:^80}".format("Starting a Bokeh release BUILD"))

for step in BUILD_CHECKS:
    step(config)

for step in BUILD_STEPS:
    step(config)

banner(blue, "{:^80}".format(f"Bokeh {config.version!r} BUILD: SUCCESS"))



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------

# Standard library imports
import os
from typing import Callable, Tuple, Union

# External imports
import boto
import boto.s3.connection
import boto.s3.key

# Bokeh imports
from .config import Config, StepType
from .system import run
from .ui import failed, passed

__all__ = (
    "verify_anaconda_credentials",
    "verify_aws_credentials",
    "verify_npm_credentials",
    "verify_pypi_credentials",
)

ReturnStepType = Callable[[Config], Union[str, Tuple[str, str]]]


def collect_credential(func: ReturnStepType) -> StepType:
    service = func.__name__.split("_")[1].upper()

    def wrapper(config: Config) -> None:
        try:
            config.credentials[service] = func(config)
            passed(f"Verified {service} credentials")
        except Exception:
            failed(f"Could NOT verify {service} credentials")
            config.abort()

    return wrapper


@collect_credential
def verify_anaconda_credentials(config: Config) -> str:
    token = os.environ["ANACONDA_TOKEN"]
    out = run(f"anaconda -t {token} whoami", silent=True)
    if "Anonymous User" in out:
        raise ValueError()
    return token


@collect_credential
def verify_pypi_credentials(config: Config) -> str:
    # TODO is there a way to actually test that the creds work?
    token = os.environ["PYPI_TOKEN"]
    return token


@collect_credential
def verify_npm_credentials(config: Config) -> str:
    token = os.environ["NPM_TOKEN"]
    run("npm set registry 'https://registry.npmjs.org'")
    run(f"npm set //registry.npmjs.org/:_authToken {token}", silent=True)
    out = run(f"npm whoami")
    if out.strip() != "bokeh":
        raise ValueError()

    return token


@collect_credential
def verify_aws_credentials(config: Config) -> Tuple[str, str]:
    access_key_id = os.environ["AWS_ACCESS_KEY_ID"]
    secret_access_key = os.environ["AWS_SECRET_ACCESS_KEY"]

    for bucket_name, bucket_region in [("cdn.bokeh.org", "us-east-1"), ("cdn-backup.bokeh.org", "us-west-2")]:
        conn = boto.s3.connect_to_region(
            bucket_region,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            calling_format=boto.s3.connection.OrdinaryCallingFormat(),
        )
        conn.get_bucket(bucket_name)

    return (access_key_id, secret_access_key)



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------

# Standard library imports
import re
from collections import defaultdict
from typing import Any, Callable, Dict

# Bokeh imports
from .enums import ActionStatus, VersionType
from .system import run
from .ui import banner, bright, red, yellow

__all__ = (
    "Config",
    "StepType",
)

# This excludes "local" build versions, e.g. 0.12.4+19.gf85560a
ANY_VERSION = re.compile(r"^(\d+\.\d+\.\d+)((dev|rc)(\d+))?$")

FULL_VERSION = re.compile(r"^(\d+\.\d+\.\d+)?$")


class Config(object):
    def __init__(self, version: str) -> None:
        if not ANY_VERSION.match(version):
            raise ValueError(f"Invalid Bokeh version for build/release {version!r}")
        self.version: str = version
        self.credentials: Dict[str, Any] = {}
        self._last_any_version: str = ""
        self._last_full_version: str = ""

        self.builds = ("conda", "sdist", "docs", "npm")
        self.build_status: Dict[str, ActionStatus] = defaultdict(lambda: ActionStatus.NOT_STARTED)

        self.uploads = ("cdn", "anaconda", "pypi", "docs", "npm")
        self.upload_status: Dict[str, ActionStatus] = defaultdict(lambda: ActionStatus.NOT_STARTED)

    def _to_js_version(self, v: str) -> str:
        if FULL_VERSION.match(v):
            return v
        match = ANY_VERSION.match(v)
        if not match:
            raise ValueError(f"Invalid verison {v!r}")
        (version, suffix, release, number) = match.groups()
        return f"{version}-{release}.{number}"

    @property
    def version_type(self) -> VersionType:
        if "rc" in self.version:
            return VersionType.RC
        elif "dev" in self.version:
            return VersionType.DEV
        else:
            return VersionType.FULL

    @property
    def repo_top_dir(self) -> str:
        return run("git rev-parse --show-toplevel").strip()

    @property
    def js_version(self) -> str:
        return self._to_js_version(self.version)

    @property
    def last_any_version(self) -> str:
        return self._last_any_version

    @last_any_version.setter
    def last_any_version(self, v: str) -> None:
        m = ANY_VERSION.match(v)
        if not m:
            raise ValueError("Invalid Bokeh version %r" % v)
        self._last_any_version = v

    @property
    def js_last_any_version(self) -> str:
        return self._to_js_version(self.last_any_version)

    @property
    def last_full_version(self) -> str:
        return self._last_full_version

    @last_full_version.setter
    def last_full_version(self, v: str) -> None:
        m = FULL_VERSION.match(v)
        if not m:
            raise ValueError("Invalid Bokeh version %r" % v)
        self._last_full_version = v

    @property
    def release_branch(self) -> str:
        return "release-%s" % self.version

    def abort(self) -> None:
        print()
        print(bright(red("!!! Steps failed. The BUILD has been aborted.")))
        print()

        print(bright(yellow("Here is the status of all build steps:")))
        print()
        for build in self.builds:
            print(f"    - {build:>10}: {self.build_status[build].value}")

        print()

        if all(self.upload_status[x] == ActionStatus.NOT_STARTED for x in self.uploads):
            print(bright(red("!!! NO ASSETS HAVE BEEN UPLOADED")))
        else:
            print(bright(red("!!! SOME ASSETS MAY HAVE BEEN UPLOADED")))
            print
            print(bright(yellow("Here is the status of all uploads:")))
            for upload in self.uploads:
                print(f"    - {upload:>10}: {self.upload_status[upload].value}")

        banner(red, "{:^80}".format(f"Bokeh {self.version!r} build: FAILURE"))
        raise RuntimeError()
        # sys.exit(1)


StepType = Callable[[Config], None]



# -----------------------------------------------------------------------------
# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------

# Standard library imports
import os
from subprocess import PIPE, STDOUT
from subprocess import run as stdlib_run
from typing import Any, Optional

__all__ = (
    "cd",
    "run",
)


def run(cmd: str, fake_cmd: Optional[str] = None, silent: bool = True, **kw: Any) -> str:

    if not silent:
        envstr = " ".join(f"{k}={v}" for k, v in kw.items())
        if envstr:
            envstr += " "
        if fake_cmd:
            print(f"+{envstr}{fake_cmd}")
        else:
            print(f"+{envstr}{cmd}")

    env = dict(os.environ)
    env.update(kw)

    result = stdlib_run(cmd, shell=True, stdout=PIPE, stderr=STDOUT, text=True, env=env)  # type: ignore

    result.check_returncode()

    return result.stdout or ""


def cd(dir: str, silent: bool = True) -> None:
    os.chdir(dir)
    if not silent:
        print("+cd %s    [now: %s]" % (dir, os.getcwd()))
